import warnings
import datetime as dt
from astropy.time import Time


def parse_str(input_string, metadata):
    """Parse string input.

    Args:
        input_string (str): String to parse.
        metadata (MetaDataSection): Metadata corresponding to this string.

    Returns:
        parsed_str (str): Parsed string.
    """
    return str(input_string)


def parse_datetime(epoch):
    """Convert OEM standard epoch to a DateTime.

    Args:
        epoch (str): OEM epoch string.

    Returns:
        parsed_epoch (DateTime):  Parsed epoch.
    """
    if "." in epoch:
        return dt.datetime.strptime(
            epoch.replace("Z", "")[:epoch.index(".")+7].strip(),
            "%Y-%m-%dT%H:%M:%S.%f"
        )
    else:
        return dt.datetime.strptime(
            epoch.replace("Z", "").strip(),
            "%Y-%m-%dT%H:%M:%S"
        )


def parse_utc(epoch, metadata):
    """Parse OEM standard UTC epoch.

    Args:
        epoch (str): OEM epoch string.

    Returns:
        parsed_epoch (Time): UTC epoch.
    """
    return Time(epoch, format="isot", scale="utc")


def parse_epoch(epoch, metadata):
    """Parse OEM standard epoch using metadata TIME_SYSTEM.

    Args:
        epoch (str): OEM epoch string.
        metadata (MetaDataSection): Metadata corresponding to this epoch.

    Returns:
        parsed_epoch (Time): Parsed epoch with assigned time scale. If the
            timescale indicated by TIME_SYSTEM is not supported by astropy,
            then parsed_epoch will warn the user and fall back to DateTime. In
            this case, time calculations may be inaccurate.
    """
    time_system = metadata["TIME_SYSTEM"].lower()
    if time_system in Time.SCALES:
        parsed_epoch = Time(epoch, format="isot", scale=time_system)
    else:
        warnings.warn(
            f"Unsupported TIME_SYSTEM '{time_system}', falling back to "
            f"DateTime. Use caution with time calculations."
        )
        parsed_epoch = parse_datetime(epoch)
    return parsed_epoch


def parse_integer(input, metadata):
    """Parse integer value.

    Args:
        input: Any input value that can be cast as a number.

    Returns:
        integer (int): Integer equivalent of input.

    Raises:
        ValueError: Invalid integer.
    """
    if float(input).is_integer():
        return int(input)
    else:
        raise ValueError(f"Invalid integer: '{input}'")


def format_float(value):
    """Convert float to a common string format.

    Args:
        value: Any input that can be cast as a float.

    Returns:
        formatted_value (str): Float following standard format.
    """
    return f"{value:+.14e}"


def format_epoch(epoch):
    """Format an epoch in the standard OEM format.

    Args:
        epoch (Time, DateTime): Epoch to convert to string.

    Returns:
        formatted_epoch (str): Epoch in YYYY-MM-DDTHH:MM:SS.sss format.
    """
    return epoch.strftime("%Y-%m-%dT%H:%M:%S.%f")


def require(boolean, message):
    """Require a boolean condition.

    Args:
        boolean (bool): Condition boolean.
        message (str): Error message.

    Raises:
        ValueError: message
    """
    if not boolean:
        raise ValueError(message)


def require_field(field, metadata):
    """Require a field in a dict.

    Args:
        field (str): String containing the required field.
        metadata (dict): Dictionary to check for key.

    Raises:
        ValueError: Missing required header.
    """
    require(field in metadata, f"Missing required header: {field}")


def is_kvn(file_path):
    """Determine if an OEM file is KVN or XML.

    Args:
        file_path (str or Path): Path of file to check.

    returns:
        result (bool): True if file is KVN, false if XML.
    """
    with open(file_path, "r") as target_file:
        if "<?xml" in target_file.readline():
            result = False
        else:
            result = True
    return result



import re

import numpy as np
from lxml.etree import SubElement

from oem import patterns, CURRENT_VERSION
from oem.tools import parse_epoch, require, format_float, format_epoch
from oem.base import ConstraintSpecification, Constraint


COV_XML_ENTRY_MAP = {
    "CX_X": (0, 0),
    "CY_X": (1, 0),
    "CZ_X": (2, 0),
    "CX_DOT_X": (3, 0),
    "CY_DOT_X": (4, 0),
    "CZ_DOT_X": (5, 0),
    "CY_Y": (1, 1),
    "CZ_Y": (2, 1),
    "CX_DOT_Y": (3, 1),
    "CY_DOT_Y": (4, 1),
    "CZ_DOT_Y": (5, 1),
    "CZ_Z": (2, 2),
    "CX_DOT_Z": (3, 2),
    "CY_DOT_Z": (4, 2),
    "CZ_DOT_Z": (5, 2),
    "CX_DOT_X_DOT": (3, 3),
    "CY_DOT_X_DOT": (4, 3),
    "CZ_DOT_X_DOT": (5, 3),
    "CY_DOT_Y_DOT": (4, 4),
    "CZ_DOT_Y_DOT": (5, 4),
    "CZ_DOT_Z_DOT": (5, 5)
}


class ConstrainStateType(Constraint):

    versions = ["1.0"]

    def func(self, state):
        require(
            state.acceleration is None,
            "State in v1.0 OEM cannot have acceleration entries"
        )


class ConstrainStateDimension(Constraint):

    versions = ["1.0", "2.0"]

    def func(self, state):
        require(state.position.size == 3, "State position size != 3")
        require(state.velocity.size == 3, "State velocity size != 3")
        if state.acceleration is not None:
            require(
                state.acceleration.size == 3,
                "State acceleration size != 3"
            )


class State(object):
    """Basic Cartesian state.

    Attributes:
        epoch (DateTime): Epoch date and time.
        position (ndarray): 3-element array describing the position at epoch.
        velocity (ndarray): 3-element array describing the velocity at epoch.
        acceleration (ndarray): 3-element array describing the acceleration
            at epoch. If unavailable, this attribute is None.
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainStateType,
        ConstrainStateDimension
    )

    def __init__(self, epoch, position, velocity, acceleration=None,
                 version=CURRENT_VERSION):
        self.version = version
        self.epoch = epoch
        self.position = np.array(position)
        self.velocity = np.array(velocity)
        self.acceleration = (
            np.array(acceleration)
            if acceleration is not None else None
        )
        self._constraint_spec.apply(self)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.epoch == other.epoch and
            (self.position == other.position).all() and
            (self.velocity == other.velocity).all() and
            np.array([self.acceleration == other.acceleration]).all()
        )

    @classmethod
    def _from_string(cls, segment, version, metadata):
        """Create State from OEM-formatted string.

        Args:
            segment (str): String containing a single OEM state line.

        Returns:
            new_state (State): New State instance.
        """
        raw_state = segment.split()
        has_accel = True if len(raw_state) == 10 else False

        epoch = parse_epoch(raw_state[0], metadata)
        position = [float(entry) for entry in raw_state[1:4]]
        velocity = [float(entry) for entry in raw_state[4:7]]
        if has_accel:
            acceleration = [float(entry) for entry in raw_state[7:]]
        else:
            acceleration = None

        return cls(
            epoch,
            position,
            velocity,
            acceleration=acceleration,
            version=version
        )

    @classmethod
    def _from_xml(cls, segment, version, metadata):
        epoch = parse_epoch(segment[0].text, metadata)
        position = [float(entry.text) for entry in segment[1:4]]
        velocity = [float(entry.text) for entry in segment[4:7]]
        if len(segment) == 10:
            acceleration = [float(entry.text) for entry in segment[7:]]
        else:
            acceleration = None
        return cls(
            epoch,
            position,
            velocity,
            acceleration=acceleration,
            version=version
        )

    def _to_string(self):
        entries = list(self.position) + list(self.velocity)
        if self.has_accel:
            entries += list(self.acceleration)
        formatted_epoch = format_epoch(self.epoch)
        formatted_entries = "  ".join(
            [format_float(entry) for entry in entries]
        )
        return f"{formatted_epoch}  {formatted_entries}\n"

    def _to_xml(self, parent):
        SubElement(parent, "EPOCH").text = format_epoch(self.epoch)
        SubElement(parent, "X").text = format_float(self.position[0])
        SubElement(parent, "Y").text = format_float(self.position[1])
        SubElement(parent, "Z").text = format_float(self.position[2])
        SubElement(parent, "X_DOT").text = format_float(self.velocity[0])
        SubElement(parent, "Y_DOT").text = format_float(self.velocity[1])
        SubElement(parent, "Z_DOT").text = format_float(self.velocity[2])
        if self.has_accel:
            SubElement(parent, "X_DDOT").text = (
                format_float(self.acceleration[0]))
            SubElement(parent, "Y_DDOT").text = (
                format_float(self.acceleration[1]))
            SubElement(parent, "Z_DDOT").text = (
                format_float(self.acceleration[2]))

    @property
    def has_accel(self):
        return True if self.acceleration is not None else False


class Covariance(object):
    """Basic 6x6 covariance.

    Attributes:
        epoch (DateTime): Epoch date and time.
        frame (str): Reference from of this covariance.
        matrix (ndarray): 6x6 covariance matrix.
    """

    def __init__(self, epoch, frame, matrix, version=CURRENT_VERSION):
        self.version = version
        self.epoch = epoch
        self.frame = frame
        self.matrix = np.array(matrix)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.epoch == other.epoch and
            self.frame == other.frame and
            (self.matrix == other.matrix).all()
        )

    @classmethod
    def _from_string(cls, segment, version, metadata):
        """Create Covariance from OEM-formatted string.

        Args:
            segment (str): String containing a single OEM covariance block.

        Returns:
            new_covariance (Covariance): New Covariance instance.
        """
        headers = {
            entry[0]: entry[1].strip()
            for entry in re.findall(patterns.KEY_VAL, segment, re.MULTILINE)
        }
        if "EPOCH" not in headers:
            raise ValueError("Covariance entry missing keyword 'EPOCH'")
        else:
            epoch = parse_epoch(headers["EPOCH"], metadata)
        frame = headers.get("COV_REF_FRAME")

        raw_covariance = re.findall(
            patterns.COVARIANCE_MATRIX,
            segment,
            re.MULTILINE
        )[0]

        matrix = np.zeros((6, 6))
        covariance_lines = [
            entry for entry in raw_covariance.splitlines()
            if entry.strip()
        ]
        for row_idx, row in enumerate(covariance_lines):
            for col_idx, entry in enumerate(row.split()):
                matrix[row_idx, col_idx] = float(entry)
                matrix[col_idx, row_idx] = float(entry)

        return cls(epoch, frame, matrix, version=version)

    @classmethod
    def _from_xml(cls, segment, version, metadata):
        parts = [entry for entry in segment if entry.tag != "COMMENT"]
        entries = {entry.tag: entry.text for entry in parts}
        if "EPOCH" not in entries:
            raise ValueError("Covariance entry missing keyword 'EPOCH'")
        else:
            epoch = parse_epoch(entries["EPOCH"], metadata)
        frame = entries.get("COV_REF_FRAME")

        matrix = np.zeros((6, 6))
        for key, index in COV_XML_ENTRY_MAP.items():
            matrix[index] = float(entries[key])
            if index[0] != index[1]:
                matrix[index[::-1]] = float(entries[key])

        return cls(epoch, frame, matrix, version=version)

    def _to_string(self):
        lines = f"EPOCH = {format_epoch(self.epoch)}\n"
        if self.frame:
            lines += f"COV_REF_FRAME = {self.frame}\n"

        idx = 1
        for row in self.matrix:
            formatted_entries = [format_float(entry) for entry in row[:idx]]
            lines += "  ".join(formatted_entries) + "\n"
            idx += 1

        return lines

    def _to_xml(self, parent):
        SubElement(parent, "EPOCH").text = format_epoch(self.epoch)
        if self.frame:
            SubElement(parent, "COV_REF_FRAME").text = self.frame
        for key, index in COV_XML_ENTRY_MAP.items():
            SubElement(parent, key).text = format_float(self.matrix[index])



"""Regex patterns for parsing CCSDS ephemerides
"""


HS = r"(?:[ \t]|$)+"
"""Arbitrary Horizontal spacing or EOL"""

NL = r"(?:\n)+"
"""Arbitrary number of newlines"""

HSNL = r"(?:(?:[ \t]*\n)|[ \t]*$)+"
"""Arbitrary horizontal spacing and newlines"""

FLOAT = f"(?:[+-]?\\d\\.\\d+(?:[eE][+-]\\d{{1,3}})){HS}|[+-]?\\d+\\.\\d+"
"""Floating-point number"""

DATE = r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d+)?Z?"
"""Date and time"""

KEY_VAL = f"([A-Z_]+){HS}={HS}(.+)"
"""Key-value pair"""

KEY_VAL_NC = f"[A-Z_]+{HS}={HS}.+"
"""Non-capturing key-value pair"""

DATA_LINE = f"{DATE}{HS}(?:(?:(?:{FLOAT}){HS}){{3}}){{2,3}}"
"""Ephemeris data line"""

COMMENT_LINE = r"COMMENT.+"
"""Comment line"""

COMMENT = f"(?:{COMMENT_LINE}{HSNL})+"
"""Single comment"""

HEADER_SEGMENT = (
    f"\\ACCSDS_OEM_VERS{HS}={HS}.+{HSNL}"
    f"(?:{COMMENT})?"
    f"(?:(?:{KEY_VAL_NC}{HSNL})+)"
)

METADATA_SEGMENT = (
    f"META_START{HSNL}"
    f"(?:{COMMENT})?"
    f"(?:{KEY_VAL_NC}{HSNL})+"
    f"META_STOP{HSNL}"
)

DATA_SEGMENT = (
    f"(?:{COMMENT})?"
    f"(?:{DATA_LINE}(?:{HSNL}|$))+"
)

COVARIANCE_MATRIX = "".join([
    f"(?:{FLOAT}{HS}){{{idx}}}{HSNL}"
    for idx in range(1, 7)
])

COVARIANCE_ENTRY = (
    f"(?:{KEY_VAL_NC}{HSNL}){{1,2}}"
    f"{COVARIANCE_MATRIX}"
)

COVARIANCE_SEGMENT = (
    f"COVARIANCE_START{HSNL}"
    f"(?:{COVARIANCE_ENTRY})+"
    f"COVARIANCE_STOP(?:{HSNL}|$)"
)

DATA_BLOCK = (
    f"({METADATA_SEGMENT})"
    f"({DATA_SEGMENT})"
    f"({COVARIANCE_SEGMENT})?"
)

CCSDS_EPHEMERIS = (
    f"({HEADER_SEGMENT})"
    f"(?:{DATA_BLOCK})+\\Z"
)



import re

from lxml.etree import ElementTree, Element, SubElement, parse

from oem import components, patterns
from oem.base import Constraint, ConstraintSpecification
from oem.tools import require, is_kvn


class ConstrainOemTimeSystem(Constraint):
    """Apply constraints to OEM TIME_SYSTEM."""

    versions = ["1.0", "2.0"]

    def func(self, oem):
        time_system = None
        for segment in oem:
            if time_system is None:
                time_system = segment.metadata["TIME_SYSTEM"]
            else:
                require(
                    segment.metadata["TIME_SYSTEM"] == time_system,
                    "TIME_SYSTEM not fixed in OEM"
                )


class ConstrainOemStates(Constraint):
    '''Apply constraints to OEM data sections'''

    versions = ["1.0", "2.0"]

    def func(self, oem):
        if oem.version == "1.0":
            self.v1_0(oem)
        else:
            self.v2_0(oem)

    def v1_0(self, oem):
        require(
            all(
                (oem._segments[idx].metadata["STOP_TIME"]
                 <= oem._segments[idx+1].metadata["START_TIME"])
                for idx in range(len(oem._segments)-1)
            ),
            "Data section state epochs overlap"
        )

    def v2_0(self, oem):
        require(
            all(
                (oem._segments[idx].useable_stop_time
                 <= oem._segments[idx+1].useable_start_time)
                for idx in range(len(oem._segments)-1)
            ),
            "Data section state epochs overlap"
        )


class OrbitEphemerisMessage(object):
    """Python representation of an Orbit Ephemeris Message.

    This class provides the primary interface between the OEM module and an
    OEM file.

    Attributes:
        header (HeaderSection): Object containing the OEM header section.

    Examples:
        The `OrbitEphemerisMessage` class can load directly from a file:

        >>> ephemeris = OrbitEphemerisMessage.open(file_path)

        An OEM is made up of one or more data segments available through an
        iterator:

        >>> for segment in ephemeris:
        ...     for state in segment:
        ...         # Iterate through states
        ...         pass
        ...     for covariance in segment.covariances:
        ...         # Iterate through covariances
        ...         pass

        It is also possible to iterate through the states and covariances in
        all segments with the `.states` and `.covariances` properties.

        To determine if a particular epoch is contained in the useable time
        range of any of the segments in an ephemeris, use `in`:

        >>> epoch in ephemeris
        True

        The `save_as` method enables saving of copies of an OEM in both KVN and
        XML formats.

        >>> oem.save_as("new.oem", file_format="xml")

        To convert directly between KVN and XML formats, use the `convert`
        class method. For example, to convert a KVN OEM to XML:

        >>> oem.convert("input.oem", "output.oem", "xml")
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainOemTimeSystem,
        ConstrainOemStates
    )

    def __init__(self, header, segments):
        '''Create an Orbit Ephemeris Message.

        Args:
            header (HeaderSection): Object containing the OEM header section.
            segments (list): List of OEM EphemerisSegments.
        '''
        self.header = header
        self.version = self.header["CCSDS_OEM_VERS"]
        self._segments = segments
        self._constraint_spec.apply(self)

    def __iter__(self):
        return iter(self._segments)

    def __contains__(self, epoch):
        return any(epoch in segment for segment in self._segments)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.header == other.header and
            len(self._segments) == len(other._segments) and
            all(
                this_segment == other_segment
                for this_segment, other_segment
                in zip(self._segments, other._segments)
            )
        )

    @classmethod
    def from_kvn_oem(cls, file_path):
        with open(file_path, "r") as ephem_file:
            contents = ephem_file.read()
        contents = re.sub(patterns.COMMENT_LINE, "", contents)
        match = re.match(patterns.CCSDS_EPHEMERIS, contents, re.MULTILINE)
        if match:
            header = components.HeaderSection._from_string(match.group(1))
            version = header["CCSDS_OEM_VERS"]
            segments = [
                components.EphemerisSegment._from_strings(raw_segment, version)
                for raw_segment
                in re.findall(patterns.DATA_BLOCK, contents, re.MULTILINE)
            ]
        else:
            raise ValueError("Failed to parse ephemeris file.")
        return cls(header, segments)

    @classmethod
    def from_xml_oem(cls, file_path):
        parts = parse(str(file_path)).getroot()
        header = components.HeaderSection._from_xml(parts)
        segments = [
            components.EphemerisSegment._from_xml(part, header.version)
            for part in parts[1]
        ]
        return cls(header, segments)

    @classmethod
    def open(cls, file_path):
        """Open an Orbit Ephemeris Message file.

        This method supports both KVN and XML formats.

        Args:
            file_path (str or Path): Path of file to read.

        Returns:
            oem: OrbitEphemerisMessage instance.
        """
        if is_kvn(file_path):
            oem = cls.from_kvn_oem(file_path)
        else:
            oem = cls.from_xml_oem(file_path)
        return oem

    @classmethod
    def convert(cls, in_file_path, out_file_path, file_format):
        """Convert an OEM to a particular file format.

        This method will succeed and produce an output file even if the input
        file is already in the desired format. Comments are not preserved.

        Args:
            in_file_path (str or Path): Path to original ephemeris.
            out_file_path (str or Path): Desired path for converted ephemeris.
            file_format (str): Desired output format. Options are
                'kvn' and 'xml'.
        """
        cls.open(in_file_path).save_as(out_file_path, file_format=file_format)

    def save_as(self, file_path, file_format="kvn"):
        """Write OEM to file.

        Args:
            file_path (str or Path): Desired path for output ephemeris.
            file_format (str, optional): Type of file to output. Options are
                'kvn' and 'xml'. Default is 'kvn'.
        """
        if file_format == "kvn":
            with open(file_path, "w") as output_file:
                output_file.write(self._to_kvn_oem())
        elif file_format == "xml":
            self._to_xml_oem().write(
                str(file_path),
                pretty_print=True,
                encoding="utf-8",
                xml_declaration=True
            )
        else:
            raise ValueError(f"Unrecognized file type: '{file_format}'")

    def _to_kvn_oem(self):
        lines = self.header._to_string() + "\n"
        lines += "".join(entry._to_string() for entry in self._segments)
        return lines

    def _to_xml_oem(self):
        oem = Element("oem", id="CCSDS_OEM_VERS", version=self.version)
        self.header._to_xml(SubElement(oem, "header"))
        body = SubElement(oem, "body")
        for entry in self._segments:
            entry._to_xml(SubElement(body, "segment"))
        return ElementTree(oem)

    @property
    def states(self):
        '''Return a list of states in all segments.'''
        return [
            state
            for segment in self
            for state in segment.states
        ]

    @property
    def covariances(self):
        '''Return a list of covariances in all segments.'''
        return [
            covariance
            for segment in self
            for covariance in segment.covariances
        ]



import re

from lxml.etree import SubElement

from oem import patterns, CURRENT_VERSION
from oem.base import ConstraintSpecification, Constraint
from oem.tools import require
from oem.components.types import Covariance


class ConstrainCovarianceSectionEpochOrder(Constraint):
    """Apply constraints to data section state epochs."""

    versions = ["1.0", "2.0"]

    def func(self, covariance_section):
        require(
            all(
                (covariance_section.covariances[idx].epoch
                 < covariance_section.covariances[idx+1].epoch)
                for idx in range(len(covariance_section.covariances)-1)
            ),
            "States in data section are not ordered by epoch"
        )


class CovarianceSection(object):
    """OEM covariance section.

    Container for a single OEM covariance section.
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainCovarianceSectionEpochOrder
    )

    def __init__(self, covariances, version=CURRENT_VERSION):
        self.version = version
        self._covariances = covariances
        self._constraint_spec.apply(self)

    def __iter__(self):
        return iter(self.covariances)

    def __eq__(self, other):
        return (
            self.version == other.version and
            all(
                this_covariance == other_covariance
                for this_covariance, other_covariance
                in zip(self._covariances, other._covariances)
            )
        )

    @classmethod
    def _from_string(cls, segment, version, metadata):
        """Create CovarianceSection from OEM-formatted string.

        Args:
            segment (str): String containing a single OEM covariance section.

        Returns:
            new_section (CovarianceSection): New CovarianceSection instance.
        """
        raw_covariances = re.findall(
            patterns.COVARIANCE_ENTRY,
            segment,
            re.MULTILINE
        )
        covariances = [
            Covariance._from_string(entry, version, metadata)
            for entry in raw_covariances
        ]
        return cls(covariances, version=version)

    @classmethod
    def _from_xml(cls, segment, version, metadata):
        covariances = [
            Covariance._from_xml(entry, version, metadata)
            for entry in segment
        ]
        return cls(covariances, version)

    def _to_string(self):
        lines = "COVARIANCE_START\n"
        lines += "\n".join(
            entry._to_string()
            for entry in self._covariances
        )
        lines += "COVARIANCE_STOP\n"
        return lines

    def _to_xml(self, parent):
        for entry in self._covariances:
            entry._to_xml(SubElement(parent, "covarianceMatrix"))

    @property
    def covariances(self):
        """Return a list of covariances in this section."""
        return self._covariances



import re

from lxml.etree import SubElement

from oem import patterns, CURRENT_VERSION
from oem.base import ConstraintSpecification, Constraint
from oem.tools import require
from oem.components.types import State


class ConstrainDataSectionEpochOrder(Constraint):
    """Apply constraints to data section state epochs."""

    versions = ["1.0", "2.0"]

    def func(self, data_section):
        require(
            all(
                (data_section.states[idx].epoch
                 < data_section.states[idx+1].epoch)
                for idx in range(len(data_section.states)-1)
            ),
            "States in data section are not ordered by epoch"
        )


class ConstrainDataSectionStates(Constraint):
    """Apply constraint to data section states."""

    versions = ["1.0", "2.0"]

    def func(self, data_section):
        for state in data_section:
            if (not data_section.has_accel
                    and state.acceleration is not None
                    or data_section.has_accel
                    and state.acceleration is None):
                raise ValueError("Cannot change state type mid-segment.")


class DataSection(object):
    """OEM data section.

    Container for a single OEM ephemeris state data section.
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainDataSectionStates,
        ConstrainDataSectionEpochOrder
    )

    def __init__(self, states, version=CURRENT_VERSION):
        self.version = version
        self._has_accel = False if states[0].acceleration is None else True
        self._states = states
        self._constraint_spec.apply(self)

    def __iter__(self):
        return iter(self.states)

    def __eq__(self, other):
        return (
            self.version == other.version and
            all(
                this_state == other_state
                for this_state, other_state
                in zip(self._states, other._states)
            )
        )

    @classmethod
    def _from_string(cls, segment, version, metadata):
        """Create DataSection from OEM-formatted string.

        Args:
            segment (str): String containing a single OEM data section.

        Returns:
            new_section (DataSection): New DataSection instance.
        """
        raw_states = re.findall(patterns.DATA_LINE, segment, re.MULTILINE)
        states = [
            State._from_string(entry, version, metadata)
            for entry in raw_states
        ]
        return cls(states, version=version)

    @classmethod
    def _from_xml(cls, segment, version, metadata):
        states = [
            State._from_xml(entry, version, metadata)
            for entry in segment
            if entry.tag != "COMMENT"
        ]
        return cls(states, version=version)

    def _to_string(self):
        return "".join(entry._to_string() for entry in self._states)

    def _to_xml(self, parent):
        for entry in self._states:
            entry._to_xml(SubElement(parent, "stateVector"))

    @property
    def states(self):
        """Return a list of States in this section."""
        return self._states

    @property
    def has_accel(self):
        """Evaluate if section contains acceleration data."""
        return self._has_accel



from lxml.etree import SubElement

from oem import CURRENT_VERSION
from oem.base import ConstraintSpecification, Constraint
from oem.tools import require
from oem.components.metadata import MetaDataSection
from oem.components.data import DataSection
from oem.components.covariance import CovarianceSection


class ConstrainEphemerisSegmentCovariance(Constraint):
    '''Apply constraints to ephemeris segment covariance sections'''

    versions = ["1.0"]

    def func(self, ephemeris_segment):
        require(
            ephemeris_segment._covariance_data is None,
            "Covariance data block not supported in OEM v1.0"
        )


class ConstrainEphemerisSegmentCovarianceEpochs(Constraint):
    '''Apply constraints to ephemeris segment covariance sections'''

    versions = ["1.0", "2.0"]

    def func(self, ephemeris_segment):
        for covariance in ephemeris_segment.covariances:
            require(
                (
                    covariance.epoch >=
                    ephemeris_segment.metadata["START_TIME"]
                    and
                    covariance.epoch <=
                    ephemeris_segment.metadata["STOP_TIME"]
                ),
                f"Covariance epoch not within range: {covariance.epoch}"
            )


class ConstrainEphemerisSegmentStateEpochs(Constraint):
    '''Apply constraints to ephemeris segment state sections'''

    versions = ["1.0", "2.0"]

    def func(self, ephemeris_segment):
        for state in ephemeris_segment.states:
            require(
                (
                    state.epoch >=
                    ephemeris_segment.metadata["START_TIME"]
                    and
                    state.epoch <=
                    ephemeris_segment.metadata["STOP_TIME"]
                ),
                f"State epoch not within usable range: {state.epoch}"
            )


class EphemerisSegment(object):
    """OEM ephemeris segment.

    Container for a single OEM ephemeris segment.
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainEphemerisSegmentStateEpochs,
        ConstrainEphemerisSegmentCovariance,
        ConstrainEphemerisSegmentCovarianceEpochs
    )

    def __init__(self, metadata, state_data, covariance_data=None,
                 version=CURRENT_VERSION):
        self.version = version
        self.metadata = metadata
        self._state_data = state_data
        self._covariance_data = covariance_data
        self._constraint_spec.apply(self)

    def __contains__(self, epoch):
        return (
            epoch >= self.useable_start_time and
            epoch <= self.useable_stop_time
        )

    def __iter__(self):
        return iter(self.states)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.metadata == other.metadata and
            self._state_data == other._state_data and
            self._covariance_data == other._covariance_data
        )

    @classmethod
    def _from_strings(cls, components, version):
        """Create EphemerisSegment from OEM segment strings.

        Args:
            components (tuple): Tuple of OEM-formatted strings containing
                metadata, ephemeris data, and an optional covariance section.

        Returns:
            new_section (EphemerisSegment): New EphemerisSegment instance.
        """
        metadata = MetaDataSection._from_string(components[0], version)
        state_data = DataSection._from_string(components[1], version, metadata)
        if len(components[2]) == 0:
            covariance_data = None
        else:
            covariance_data = CovarianceSection._from_string(
                components[2], version, metadata)
        return cls(metadata, state_data, covariance_data, version=version)

    @classmethod
    def _from_xml(cls, segment, version):
        metadata = MetaDataSection._from_xml(segment[0], version)
        state_data = DataSection._from_xml(
            [
                entry for entry in segment[1]
                if entry.tag == "stateVector"
            ],
            version,
            metadata
        )

        raw_covariances = [
            entry for entry in segment[1]
            if entry.tag == "covarianceMatrix"
        ]
        if raw_covariances:
            covariance_data = CovarianceSection._from_xml(
                raw_covariances,
                version,
                metadata
            )
        else:
            covariance_data = None

        return cls(metadata, state_data, covariance_data, version=version)

    def _to_string(self):
        lines = self.metadata._to_string() + "\n"
        lines += self._state_data._to_string() + "\n"
        if self._covariance_data:
            lines += self._covariance_data._to_string() + "\n"
        return lines

    def _to_xml(self, parent):
        self.metadata._to_xml(SubElement(parent, "metadata"))
        data = SubElement(parent, "data")
        self._state_data._to_xml(data)
        if self._covariance_data:
            self._covariance_data._to_xml(data)

    @property
    def states(self):
        """Return list of States in this segment."""
        return self._state_data.states

    @property
    def covariances(self):
        """Return list of Covariances in this segment."""
        if self._covariance_data:
            return self._covariance_data.covariances
        else:
            return []

    @property
    def has_accel(self):
        """Evaluate if segment contains acceleration data."""
        return self._state_data.has_accel

    @property
    def has_covariance(self):
        """Evaluate if segment contains covariance data."""
        return True if self._covariance_data else False

    @property
    def useable_start_time(self):
        """Return epoch of start of useable state data range"""
        return self.metadata.useable_start_time

    @property
    def useable_stop_time(self):
        """Return epoch of end of useable state data range"""
        return self.metadata.useable_stop_time



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

version = os.environ.get("CI_COMMIT_TAG")[1:]
if version is None:
    raise ValueError("Version not found.")

setuptools.setup(
    name="oem",
    version=version,
    author="Brad Sease",
    author_email="bradsease@gmail.com",
    description="Python Orbit Ephemeris Message (OEM) tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bradsease/oem",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.0',
)



"""Base classes"""


class HeaderField(object):
    """Base header field specification.

    Base class for OEM Key-Value field specifications.

    Attributes:
        parser (func): Function with a single input that returns a parsed
            version of the raw header value.
        required (bool): Indication of whether or not this field is required.
    """

    def __init__(self, parser, formatter, required=False):
        self.parser = parser
        self.formatter = formatter
        self.required = required


class KeyValueSection(object):
    """Base key-value section.

    Base class for OEM Key-Value based sections.
    """

    _field_spec = {}

    def _validate_fields(self, fields):
        """Validate section fields.

        Args:
            fields (dict): Unprocessed field data.
        """
        for key in self.required_keys:
            if key not in fields:
                raise KeyError(f"Missing required header: {key}")
        for key in fields:
            if key not in self._field_spec:
                raise KeyError(f"Invalid header key: {key}")

    def _parse_fields(self, fields):
        """Parse section fields.

        Args:
            fields (dict): Unprocessed field data.
        """
        self._validate_fields(fields)
        self._fields = fields

    def _format_fields(self):
        return [f"{key} = {value}" for key, value in self._fields.items()]

    def __getitem__(self, key):
        return self._field_spec[key].parser(self._fields[key], self)

    def __setitem__(self, key, value):
        if key in self._field_spec:
            self._fields[key] = value
        else:
            raise ValueError(f"Invalid key: '{key}'")

    def __contains__(self, key):
        return key in self._fields

    def __iter__(self):
        return iter(self._fields)

    def items(self):
        return [(key, self[key]) for key in self]

    @property
    def required_keys(self):
        """Return list of keys required by this section."""
        return [
            key for key, header_spec
            in self._field_spec.items()
            if header_spec.required
        ]


class Constraint(object):
    """Base constraint type."""

    def apply(self, obj):
        """Apply constraint.

        Args:
            obj: Constrained object
        """
        if obj.version in self.versions:
            self.func(obj)


class ConstraintSpecification(object):
    """Base constraint group type."""

    def __init__(self, *constraints):
        self.constraints = constraints

    def apply(self, obj):
        """Apply all constraints in specification.

        Args:
            obj: Constrained object
        """
        for constraint in self.constraints:
            constraint().apply(obj)
